package org.jastadd.ragconnect.tests;

import read1write2.ast.*;

import java.io.IOException;
import java.util.concurrent.TimeUnit;

import static org.jastadd.ragconnect.tests.TestUtils.mqttUri;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;

/**
 * Test case "read-1-write-2".
 *
 * @author rschoene - Initial contribution
 */
public class Read1Write2Test extends AbstractMqttTest {

  private static final String TOPIC_SAME_READ = "same/read";
  private static final String TOPIC_SAME_WRITE_INT = "same/write/int";
  private static final String TOPIC_SAME_WRITE_STRING = "same/write/string";
  private static final String TOPIC_DIFFERENT_READ = "different/read";
  private static final String TOPIC_DIFFERENT_WRITE1_INT = "different/write1/int";
  private static final String TOPIC_DIFFERENT_WRITE1_STRING = "different/write1/string";
  private static final String TOPIC_DIFFERENT_WRITE2_INT = "different/write2/int";
  private static final String TOPIC_DIFFERENT_WRITE2_STRING = "different/write2/string";
  private static final String INITIAL_VALUE = "-1";

  private MqttHandler handler;
  private A model;
  private OnSameNonterminal onSameNonterminal;
  private OnDifferentNonterminal onDifferentNonterminal;
  private TheOther other1;
  private TheOther other2;

  private ReceiverData dataSame;
  private ReceiverData dataOther1;
  private ReceiverData dataOther2;

  @Override
  public void closeConnections() {
    if (handler != null) {
      handler.close();
    }
    if (model != null) {
      model.ragconnectCloseConnections();
    }
  }



  @Override
  protected void communicateSendInitialValue() throws InterruptedException {
    // check initial value
    TestUtils.waitForMqtt();
    checkData(1, Integer.parseInt(INITIAL_VALUE), prefixed(INITIAL_VALUE), 1, Integer.parseInt(INITIAL_VALUE), prefixed(INITIAL_VALUE));

    // set new value
    sendData("2", "3");

    // check new value
    TestUtils.waitForMqtt();
    checkData(2, 2, prefixed("2"), 2, 3, prefixed("3"));

    // set new value
    sendData("4", "4");

    // check new value
    TestUtils.waitForMqtt();
    checkData(3, 4, prefixed("4"), 3, 4, prefixed("4"));

    // set new value only for same
    setDataOnlySame("77");

    // check new value
    TestUtils.waitForMqtt();
    checkData(4, 77, prefixed("77"), 3, 4, prefixed("4"));
  }

  private String prefixed(String s) {
    return "prefix" + s;
  }

  @Override
  protected void communicateOnlyUpdatedValue() throws InterruptedException {
// check initial value
    TestUtils.waitForMqtt();
    checkData(0, null, null, 0, null, null);

    // set new value
    sendData("2", "3");

    // check new value
    TestUtils.waitForMqtt();
    checkData(1, 2, prefixed("2"), 1, 3, prefixed("3"));

    // set new value
    sendData("4", "4");

    // check new value
    TestUtils.waitForMqtt();
    checkData(2, 4, prefixed("4"), 2, 4, prefixed("4"));

    // set new value only for same
    setDataOnlySame("77");

    // check new value
    TestUtils.waitForMqtt();
    checkData(3, 77, prefixed("77"), 2, 4, prefixed("4"));
  }

  @Override
  protected void createModel() {
    // Setting value for Input without dependencies does not trigger any updates
    model = new A();

    onSameNonterminal = new OnSameNonterminal();
    model.setOnSameNonterminal(onSameNonterminal);
    onSameNonterminal.setInput(INITIAL_VALUE);

    onDifferentNonterminal = new OnDifferentNonterminal();
    other1 = new TheOther();
    other2 = new TheOther();
    onDifferentNonterminal.addTheOther(other1);
    onDifferentNonterminal.addTheOther(other2);
    model.setOnDifferentNonterminal(onDifferentNonterminal);
    onDifferentNonterminal.setInput(INITIAL_VALUE);
  }

  @Override
  protected void setupReceiverAndConnect(boolean writeCurrentValue) throws IOException {
    model.ragconnectSetupMqttWaitUntilReady(2, TimeUnit.SECONDS);

    handler = new MqttHandler().dontSendWelcomeMessage().setHost(TestUtils.getMqttHost());
    assertTrue(handler.waitUntilReady(2, TimeUnit.SECONDS));

    onSameNonterminal.addIntDependency(onSameNonterminal);
    onSameNonterminal.addStringDependency(onSameNonterminal);
    other1.addIntDependency(onDifferentNonterminal);
    other1.addStringDependency(onDifferentNonterminal);
    other2.addIntDependency(onDifferentNonterminal);
    other2.addStringDependency(onDifferentNonterminal);

    dataSame = new Read1Write2Test.ReceiverData();
    dataOther1 = new Read1Write2Test.ReceiverData();
    dataOther2 = new Read1Write2Test.ReceiverData();

    handler.newConnection(TOPIC_SAME_WRITE_INT, bytes -> {
      dataSame.numberOfIntValues += 1;
      dataSame.lastIntValue = java.nio.ByteBuffer.wrap(bytes).getInt();
    });
    handler.newConnection(TOPIC_SAME_WRITE_STRING, bytes -> {
      dataSame.numberOfStringValues += 1;
      dataSame.lastStringValue = new String(bytes);
    });

    handler.newConnection(TOPIC_DIFFERENT_WRITE1_INT, bytes -> {
      dataOther1.numberOfIntValues += 1;
      dataOther1.lastIntValue = java.nio.ByteBuffer.wrap(bytes).getInt();
    });
    handler.newConnection(TOPIC_DIFFERENT_WRITE1_STRING, bytes -> {
      dataOther1.numberOfStringValues += 1;
      dataOther1.lastStringValue = new String(bytes);
    });

    handler.newConnection(TOPIC_DIFFERENT_WRITE2_INT, bytes -> {
      dataOther2.numberOfIntValues += 1;
      dataOther2.lastIntValue = java.nio.ByteBuffer.wrap(bytes).getInt();
    });
    handler.newConnection(TOPIC_DIFFERENT_WRITE2_STRING, bytes -> {
      dataOther2.numberOfStringValues += 1;
      dataOther2.lastStringValue = new String(bytes);
    });

    onSameNonterminal.connectInput(mqttUri(TOPIC_SAME_READ));
    onSameNonterminal.connectOutInteger(mqttUri(TOPIC_SAME_WRITE_INT), writeCurrentValue);
    onSameNonterminal.connectOutString(mqttUri(TOPIC_SAME_WRITE_STRING), writeCurrentValue);

    onDifferentNonterminal.connectInput(mqttUri(TOPIC_DIFFERENT_READ));
    other1.connectOutInteger(mqttUri(TOPIC_DIFFERENT_WRITE1_INT), writeCurrentValue);
    other1.connectOutString(mqttUri(TOPIC_DIFFERENT_WRITE1_STRING), writeCurrentValue);
    other2.connectOutInteger(mqttUri(TOPIC_DIFFERENT_WRITE2_INT), writeCurrentValue);
    other2.connectOutString(mqttUri(TOPIC_DIFFERENT_WRITE2_STRING), writeCurrentValue);
  }

  private void sendData(String inputSame, String inputDifferent) {
    handler.publish(TOPIC_SAME_READ, inputSame.getBytes());
    handler.publish(TOPIC_DIFFERENT_READ, inputDifferent.getBytes());
  }

  private void setDataOnlySame(String inputSame) {
    handler.publish(TOPIC_SAME_READ, inputSame.getBytes());
  }

  private void checkData(int numberOfSameValues, Integer lastSameIntValue, String lastSameStringValue,
                         int numberOfDifferentValues, Integer lastDifferentIntValue,
                         String lastDifferentStringValue) {
    /* the value "-2" is never used in the test, so a test will always fail comparing to this value
     especially, it is not the initial value */
    ReceiverData expectedDataSame = ReceiverData.of(
        numberOfSameValues,
        lastSameIntValue != null ? lastSameIntValue : -2,
        lastSameStringValue);
    compareData(expectedDataSame, dataSame);
    ReceiverData expectedDataDifferent = ReceiverData.of(
        numberOfDifferentValues,
        lastDifferentIntValue != null ? lastDifferentIntValue : -2,
        lastDifferentStringValue);
    compareData(expectedDataDifferent, dataOther1);
    compareData(expectedDataDifferent, dataOther2);
  }

  private void compareData(ReceiverData expectedData,
                         ReceiverData actual) {
    assertEquals(expectedData.numberOfIntValues, actual.numberOfIntValues);
    assertEquals(expectedData.numberOfStringValues, actual.numberOfStringValues);
    if (expectedData.numberOfIntValues > 0) {
      assertEquals(expectedData.lastIntValue, actual.lastIntValue);
    }
    if (expectedData.numberOfStringValues > 0) {
      assertEquals(expectedData.lastStringValue, actual.lastStringValue);
    }
  }

  private static class ReceiverData {
    int lastIntValue;
    int numberOfIntValues = 0;
    String lastStringValue;
    int numberOfStringValues = 0;

    static ReceiverData of(int numberOfValues, int lastIntValue, String lastStringValue) {
      ReceiverData result = new ReceiverData();
      result.lastIntValue = lastIntValue;
      result.lastStringValue = lastStringValue;
      result.numberOfIntValues = numberOfValues;
      result.numberOfStringValues = numberOfValues;
      return result;
    }
  }

}
